//! error resistant stream cypher <- recover from block errors (ECC to encrypted block)

use crate::*;
use serde::{de::DeserializeOwned, Serialize};
use std::fs::{create_dir_all, read_to_string};

#[derive(PartialEq, Debug)]
pub enum CryptoError {
    /// Backend is not available or writeable
    InvalidBackend,
    /// Prevented overwrite of existing files
    BackendAlreadyExists,
    InvalidSuperBlock(String),
    NoMatingOpeningKey,
    GeneralIO(String),
    GeneralRingError,
    /// If you see this error, please update your software
    RekeyRequired,
    /// Error occured during paper key strengthening
    InvalidPassword(argon2::password_hash::Error),
}

impl From<serde_json::Error> for CryptoError {
    fn from(e: serde_json::Error) -> Self {
        CryptoError::InvalidSuperBlock(format!("{e:?}"))
    }
}
impl From<std::io::Error> for CryptoError {
    fn from(e: std::io::Error) -> Self {
        CryptoError::GeneralIO(format!("{e:?}"))
    }
}
impl From<ring::error::Unspecified> for CryptoError {
    fn from(e: ring::error::Unspecified) -> Self {
        CryptoError::GeneralRingError
    }
}
impl From<argon2::password_hash::Error> for CryptoError {
    fn from(e: argon2::password_hash::Error) -> Self {
        CryptoError::InvalidPassword(e)
    }
}

pub fn read_json_from_file<T: DeserializeOwned, P: AsRef<Path>>(path: P) -> CryptoResult<T> {
    Ok(serde_json::from_str(&read_to_string(path)?)?)
}

pub fn write_json_to_file<S: Serialize, P: AsRef<Path>>(obj: &S, path: P) -> CryptoResult<()> {
    let mut file = StdFile::open(path)?;
    Ok(serde_json::to_writer(file, obj)?)
}

pub fn path_push<B: AsRef<Path>, E: AsRef<Path>>(base: B, ext: E) -> PathBuf {
    let mut buf = base.as_ref().to_owned();
    buf.push(ext);
    buf
}

/// Generate a secure random value
pub fn random<T: ring::rand::RandomlyConstructable>() -> T {
    let rng = ring::rand::SystemRandom::new();
    ring::rand::generate(&rng)
        .expect("unable to generate random value")
        .expose()
}
pub fn random_usize() -> usize {
    let r = random::<[u8; 8]>();
    usize::from_le_bytes(r)
}

pub fn encrypt_bytes(bytes: &[u8], key: &MasterKey) -> Vec<u8> {
    todo!()
}

#[derive(Debug)]
/// Nonce length 12 Bytes
pub struct NonceCounter {
    counter: u128,
}
impl NonceCounter {
    pub fn one() -> Self {
        Self { counter: 1 }
    }
    pub fn random_lower_half() -> Self {
        Self {
            counter: random_usize() as u128,
        }
    }
}
impl ring::aead::NonceSequence for NonceCounter {
    fn advance(&mut self) -> Result<ring::aead::Nonce, ring::error::Unspecified> {
        if let None = self.counter.checked_add(1) {
            self.counter = 1;
        } else {
            self.counter += 1;
        }

        let nonce = self.counter.to_ne_bytes();
        //.to_le_bytes();

        ring::aead::Nonce::try_assume_unique_for_key(&nonce[..12])
    }
}
